#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#include <vector>
#include <iostream>
#include <string>
#include <cstdlib>
#include <sstream>
#include <libassert/assert.hpp>
#include <algorithm>
#include "execute.hpp"
#include "contrib/base64.hpp"
#include "utils/utils.hpp"
#include "utils/directories.hpp"
#include <termios.h>

bool EXITSTATUSCHECK(int ret) {
    return (ret != -1 && WIFEXITED(ret) && (WEXITSTATUS(ret) == 0)); // ret is -1 if the command failed to execute.
}

namespace dropshell {
bool __execute_command(std::vector<std::string> command, std::string * output)
{
    // Save TTY state if possible
    struct termios orig_termios;
    int tty_fd = isatty(STDIN_FILENO) ? STDIN_FILENO : -1;
    if (tty_fd != -1) {
        tcgetattr(tty_fd, &orig_termios);
    }

    std::cout << "Executing command: ";
    for (auto & x : command) std::cout << "[" << x << "] ";
    std::cout << std::endl << std::flush;

    int pipefd[2];
    bool capture = (output != nullptr);
    if (capture && pipe(pipefd) == -1) {
        perror("pipe failed");
        return false;
    }

    pid_t pid = fork();

    if (pid == -1) {
        // Fork failed
        perror("fork failed");
        return false;
    } else if (pid == 0) {
        // Child process
        if (capture) {
            close(pipefd[0]); // Close read end
            dup2(pipefd[1], STDOUT_FILENO); // Redirect stdout to pipe
            dup2(pipefd[1], STDERR_FILENO); // Redirect stderr to pipe
            close(pipefd[1]);
        }

        std::vector<const char *> commandvec;
        for (auto & x : command) {
            commandvec.push_back(x.c_str());
        }
        commandvec.push_back(NULL);

        // if (!silent) {
        //     std::cout << "Executing command: ";
        //     for (auto & x : commandvec) std::cout << x << " ";
        //     std::cout << std::endl;
        // }

        execvp(commandvec[0], const_cast<char* const*>(commandvec.data()));
        // If execvp returns, it means an error occurred
        perror("execvp failed");
        exit(EXIT_FAILURE); // Exit child process on error
    } else {
        // Parent process
        if (capture) {
            close(pipefd[1]); // Close write end
            char buffer[256];
            ssize_t count;
            while ((count = read(pipefd[0], buffer, sizeof(buffer))) > 0) {
                output->append(buffer, count);
            }
            close(pipefd[0]);
        }
        int ret;
        // Wait for the child process to complete
        waitpid(pid, &ret, 0);

        // Restore TTY state if possible
        if (tty_fd != -1) {
            tcsetattr(tty_fd, TCSANOW, &orig_termios);
        }

        return EXITSTATUSCHECK(ret);
    }
}

bool execute_command(const sSSHInfo * ssh_info, const sCommand command, cMode mode, std::string * output)
{
    if (command.get_command_to_run().empty())
        return false;
    
    std::vector<std::string> commandvec;

    if (ssh_info)
    {
        std::vector<std::string> ssh_command = {"/usr/bin/ssh", "-p", ssh_info->port, "-tt", ssh_info->user + "@" + ssh_info->host};
        commandvec.insert(commandvec.end(), ssh_command.begin(), ssh_command.end());
    }

    commandvec.push_back("bash");
    commandvec.push_back("-c");

    std::string shell_command = "";
    for (auto & x : command.get_command_to_run()) {
        shell_command += x + " ";
    }
    shell_command = shell_command.substr(0, shell_command.size() - 1);
    commandvec.push_back(shell_command);


    if (hasFlag(mode, cMode::CaptureOutput)) {
        ASSERT(output, "Capture output mode requires an output string to be provided");
        return __execute_command(commandvec, output);
    }
    else
    {
        return __execute_command(commandvec, nullptr);
    }
}

bool execute_command(const sSSHInfo ssh_info, const sCommand command, const cMode mode)
{
    return execute_command(&ssh_info, command, mode, nullptr);
}

bool execute_command(const sSSHInfo ssh_info, const sCommand command, const cMode mode, std::string & output)
{
    return execute_command(&ssh_info, command, mode, &output);
}

bool execute_command(const sCommand command, const cMode mode)
{
    return execute_command(nullptr, command, mode, nullptr);
}

bool execute_command(const sCommand command, const cMode mode, std::string & output)
{
    return execute_command(nullptr, command, mode, &output);
}


} // namespace dropshell